/*
 * This source is in the public domain, but please keep this comment.
 * Pelle Johansson
 */

#include <stdio.h>
#include <netdb.h>
#include <sys/socket.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <math.h>

#ifndef SIGPROF
#define SIGPROF 27
#endif

#ifndef NI_WITHSCOPEID
#define NI_WITHSCOPEID 0
#endif

#define BUFSIZE 200000
#define MAX_SCORE_TIME 200000
//#define USE_MAXSCORE 1

#define XY(x, y) game.field[((y) - 1) * game.ncols + ((x) - 1)]
#define SXY(x, y) game.score_field[((y) - 1) * game.ncols + ((x) - 1)]
#define NORMAL '.'
#define WALL '#'
#define LETHAL '~'
#define DANGEROUS '!'
#define BASE '@'

typedef struct scorefield
{
  char area, dir;
  int score;
#ifdef USE_MAXSCORE
  int maxscore;
#endif
} scorefield_t;

typedef struct game
{
  int ncols, nrows;
  char *field;
  scorefield_t *score_field;
} game_t;

typedef struct player
{
  int n, x, y, lx, ly;
  int score;
  int xgoal, ygoal;
  int update;
  struct player *next;
} player_t;

typedef struct package
{
  int n, w, x, y, xgoal, ygoal;
  int carrier;
  char rabid;
  float score;
  struct package *next;
} package_t;

const struct itimerval timer = {{0, 0}, {0, 950000}};
const struct itimerval timer0 = {{0, 9}, {0, 0}};

int sock;
volatile int send_now;
int max_carry, carrying = 0, money;
game_t game;
int turn = 0;
player_t *me;
player_t *players = NULL;
package_t *packages = NULL;
int all_alone = 0;
int score_field_each_turn = 1; // OBS! Always on.
char want_go;
int force, verbose = 0;
float bMWorth = 0, bMSteps = 0, pMWorth = 0;
int nBases = 0, nPkgs = 0;

int open_conn(char *host, char *port)
{
  struct addrinfo hints = {0}, *res, *curr;
  int s;
  char hbuf[NI_MAXHOST], pbuf[NI_MAXSERV];
  
  hints.ai_socktype = SOCK_STREAM;
  
  if((errno = getaddrinfo(host, port, &hints, &res)) || !res)
  {
    fprintf(stderr, "%s:%s: %s\n", host, port, gai_strerror(errno));
    return NULL;
  }
  
  for(curr = res; curr; curr = curr->ai_next)
  {
    s = socket(curr->ai_family, curr->ai_socktype, curr->ai_protocol);
    if(s < 0)
      continue;
    
    if(!connect(s, curr->ai_addr, curr->ai_addrlen))
      break;
    
    if(curr->ai_canonname)
      perror(curr->ai_canonname);
    else if(!getnameinfo(curr->ai_addr, curr->ai_addrlen, hbuf,
          sizeof(hbuf), pbuf, sizeof(pbuf), NI_NUMERICHOST | NI_NUMERICSERV |
          NI_WITHSCOPEID))
      fprintf(stderr, "%s:%s: %s.\n", hbuf, pbuf, strerror(errno));
    else
      perror("connect");
    close(s);
  }
  freeaddrinfo(res);
  
  if(!curr)
  {
    fprintf(stderr, "%s:%s: Couldn't connect.\n", host,
          port);
    return NULL;
  }
  
  printf("Connected!\n");
  return s;
}

char *get_line(char *dst)
{
  static char buf[BUFSIZE] = "";
  char *bp;
  long l;
  
  while(!(bp = strchr(buf, '\n')))
  {
    bp = &buf[strlen(buf)];
    l = read(sock, bp, BUFSIZE - strlen(buf) - 1);
    if(l < 0)
      return NULL;
    bp[l] = 0;
    if(strlen(buf) == BUFSIZE - 1)
    {
      fprintf(stderr, "Buffer overflow!\n");
      return NULL;
    }
  }
  
  *bp++ = 0;
  strcpy(dst, buf);
  strcat(dst, "\n");
  memmove(buf, bp, strlen(bp) + 1);
  return dst;
}

void put_line(const char *line)
{
  write(sock, line, strlen(line));
}

void sigprof(int unused)
{
  send_now = 1;
}

player_t *find_player(int pn)
{
  player_t *pl;
  
  if(!players)
    pl = players = malloc(sizeof(player_t));
  else
  {
    for(pl = players; pl->next; pl = pl->next)
    {
      if(pl->n == pn)
        break;
    }
    if(pl->n == pn)
    {
      pl->update = turn;
      return pl;
    }
    pl = pl->next = malloc(sizeof(player_t));
  }
  if(!pl)
    return NULL;
  
  pl->n = pn;
  pl->x = pl->y = pl->lx = pl->ly = -1;
  pl->score = 0;
  pl->xgoal = pl->ygoal = -1;
  pl->update = turn;
  pl->next = NULL;
  
  return pl;
}

package_t *find_package(int pn)
{
  package_t *pkg;
  
  if(!packages)
    pkg = packages = malloc(sizeof(package_t));
  else
  {
    for(pkg = packages; pkg->next; pkg = pkg->next)
    {
      if(pkg->n == pn)
        break;
    }
    if(pkg->n == pn)
      return pkg;
    pkg = pkg->next = malloc(sizeof(package_t));
  }
  if(!pkg)
    return NULL;
  
  pkg->n = pn;
  pkg->x = pkg->y = -1;
  pkg->w = -1;
  pkg->rabid = 0;
  pkg->carrier = -1;
  pkg->xgoal = pkg->ygoal = -1;
  pkg->next = NULL;
  
  return pkg;
}

void process_package_line(char *line)
{
  int p, xgoal, ygoal, w, n;
  package_t *pkg;
  int i = 0;
  int stepSum = 0, worthSum = 0;
  
  while((n = sscanf(line, "%d %d %d %d %[^\n]", &p, &xgoal, &ygoal, &w, line))
        >= 4)
  {
    if(n < 5)
      *line = 0;
    if(i++ == 100)
      break;
    
    if(SXY(xgoal, ygoal).area <= 1)
      stepSum += SXY(xgoal, ygoal).score;
    else
      stepSum += 999999;
    worthSum += w;
    
    pkg = find_package(p);
    if(pkg)
    {
      pkg->x = me->x;
      pkg->y = me->y;
      if(pkg->w == -1)
        pMWorth += (w - pMWorth) / ++nPkgs;
      pkg->w = w;
      pkg->xgoal = xgoal;
      pkg->ygoal = ygoal;
      pkg->carrier = -1;
    }
  }
  
  if(me && XY(me->x, me->y) == BASE)
  {
    bMSteps += (stepSum - bMSteps) / ++nBases;
    bMWorth += (worthSum - bMWorth) / nBases;
  }
}

void process_event_line(char *line)
{
  char *part;
  static char buf[BUFSIZE];
  int p;
  player_t *pl, *ppl;
  package_t *pkg, *ppkg;
  
  while((part = strsep(&line, "#")))
  {
    buf[0] = 0;
    if(sscanf(part, "%d %[^\n]", &p, buf) >= 1)
    {
      pl = find_player(p);
      if(!pl)
        continue;
      
      pl->lx = pl->x;
      pl->ly = pl->y;
      while(buf[0])
      {
        switch(buf[0])
        {
        case 'X':
          printf("\007");
          if(sscanf(buf, "X %d %[^\n]", &pl->x, buf) < 2)
            buf[0] = 0;
          break;
        case 'Y':
          if(sscanf(buf, "Y %d %[^\n]", &pl->y, buf) < 2)
            buf[0] = 0;
          break;
        case 'N':
          if(pl == me && want_go && want_go != 'N')
            force += 10;
          else if(pl == me)
            force = 1;
          if(sscanf(buf, "N %[^\n]", buf) < 1)
            buf[0] = 0;
          pl->y++;
          break;
        case 'S':
          if(pl == me && want_go && want_go != 'S')
            force += 10;
          else if(pl == me)
            force = 1;
          if(sscanf(buf, "S %[^\n]", buf) < 1)
            buf[0] = 0;
          pl->y--;
          break;
        case 'E':
          if(pl == me && want_go && want_go != 'E')
            force += 10;
          else if(pl == me)
            force = 1;
          if(sscanf(buf, "E %[^\n]", buf) < 1)
            buf[0] = 0;
          pl->x++;
          break;
        case 'W':
          if(pl == me && want_go && want_go != 'W')
            force += 10;
          else if(pl == me)
            force = 1;
          if(sscanf(buf, "W %[^\n]", buf) < 1)
            buf[0] = 0;
          pl->x--;
          break;
        case 'P':
          if(sscanf(buf, "P %d %[^\n]", &p, buf) < 2)
            buf[0] = 0;
          pkg = find_package(p);
          if(pkg)
            pkg->carrier = pl->n;
          if(pl == me)
            force = 1;
          break;
        case 'D':
          if(sscanf(buf, "D %d %[^\n]", &p, buf) < 2)
            buf[0] = 0;
          pkg = find_package(p);
          if(pkg)
          {
            pkg->x = pl->x;
            pkg->y = pl->y;
            pkg->carrier = -1;
            if(buf[0] == 0 || (pkg->x == pkg->xgoal && pkg->y == pkg->ygoal))
            {
              if(pkg->w != -1)
                pl->score += pkg->w;
              else
                pl->score++;
              if(pkg == packages)
                packages = pkg->next;
              else
              {
                for(ppkg = packages; ppkg->next != pkg; ppkg = ppkg->next)
                  ;
                ppkg->next = pkg->next;
              }
              free(pkg);
            }
          }
          if(pl == me)
            force = 1;
          break;
        default:
          buf[0] = 0;
          break;
        }
      }
    }
  }
  
  // Remove dead players.
  for(pl = players, ppl = NULL; pl; )
  {
    if(pl->update < turn || XY(pl->x, pl->y) == LETHAL)
    {
      if(pl == me)
      {
        printf("I died!\n");
        me = NULL;
      }
      for(pkg = packages, ppkg = NULL; pkg; )
      {
        if(pkg->carrier == pl->n)
        {
          if(ppkg)
            ppkg->next = pkg->next;
          else
            packages = pkg->next;
          free(pkg);
          if(ppkg)
            pkg = ppkg->next;
          else
            pkg = packages;
        }
        else
        {
          ppkg = pkg;
          pkg = pkg->next;
        }
      }
      if(ppl)
        ppl->next = pl->next;
      else
        players = pl->next;
      free(pl);
      if(ppl)
        pl = ppl->next;
      else
        pl = players;
    }
    else
    {
      ppl = pl;
      pl = pl->next;
    }
  }
  
  // Update package positions.
  for(pkg = packages; pkg;)
  {
    if(pkg->rabid)
      pkg->rabid--;
    if(pkg->carrier != -1)
    {
      pl = find_player(pkg->carrier);
      if(pl)
      {
        pkg->x = pl->x;
        pkg->y = pl->y;
      }
    }
    if(pkg->x == me->x && pkg->y == me->y && pkg->xgoal == -1)
    {
      if(pkg == packages)
        ppkg = packages = pkg->next;
      else
      {
        for(ppkg = packages; ppkg->next != pkg; ppkg = ppkg->next)
          ;
        ppkg = ppkg->next = pkg->next;
      }
      free(pkg);
      pkg = ppkg;
    }
    else
      pkg = pkg->next;
  }
  
  if(players)
    all_alone = (players->next == NULL);
  
  turn++;
}

void add_to_chain(int i, int *chain, int cstart, int cend, int score)
{
  int k;
  
  if(game.field[i] != DANGEROUS)
  {
    for(k = cend - 1; game.field[chain[k]] == DANGEROUS &&
          game.score_field[chain[k]].score > score && k >= 
          cstart; k--)
      chain[k + 1] = chain[k];
    chain[k + 1] = i;
    game.score_field[i].score = score + 3;
  }
  else
  {
    chain[cend] = i;
    if(all_alone)
      game.score_field[i].score = score + 3;
    else
      game.score_field[i].score = score + 4;
  }
}

void make_score_field(int sx, int sy)
{
  int *chain, *parents;
  int cstart, cend;
  char area;
  int i, j, x, y, si;
  scorefield_t *sf;
#if USE_MAXSCORE
  int gotdir;
#endif
  
  if(game.score_field)
    free(game.score_field);
  game.score_field = calloc(game.nrows * game.ncols, sizeof(scorefield_t));
  chain = malloc(game.nrows * game.ncols * sizeof(int));
  parents = malloc(game.nrows * game.ncols * sizeof(int));
  memset(parents, 0xFF, game.nrows * game.ncols * sizeof(int));
  
  if(sx && sy)
    si = (sy - 1) * game.ncols + sx - 1;
  else
    si = 0;
  
  area = 1;
  
  i = si;
  do
  {
    if(send_now)
      break;
    if(game.field[i] != WALL && game.field[i] != LETHAL &&
          !game.score_field[i].area)
    {
      if(area > 1 && si)
      {
        // Just realised we don't want the other areas.
        game.score_field[i].area = 2;
        if(++i == game.ncols * game.nrows)
          i = 0;
        continue;
      }
      
      cend = cstart = 0;
      chain[cend++] = i;
      
      while(cend > cstart)
      {
        if(send_now)
          break;
        
        j = chain[cstart++];
        sf = &game.score_field[j];
        
        sf->area = area;
        if(parents[j] != -1)
        {
          //sf->score = game.score_field[parents[j]].score + 1;
          sf->dir = game.score_field[parents[j]].dir;
          if(!sf->dir)
          {
            if(parents[j] == j - 1)
              sf->dir = 'E';
            else if(parents[j] == j + 1)
              sf->dir = 'W';
            else if(parents[j] == j - game.ncols)
              sf->dir = 'N';
            else if(parents[j] == j + game.ncols)
              sf->dir = 'S';
            else
              printf("%d %d\n", parents[j], j);
          }
        }
        else
          sf->score = 1;
        
#if USE_MAXSCORE
        gotdir = 0;
#endif
        x = j % game.ncols + 1;
        y = j / game.ncols + 1;
        if(x > 1  && parents[j] != j - 1 && parents[j - 1] == -1 &&
              XY(x - 1, y) != WALL && XY(x - 1, y) != LETHAL)
        {
#if USE_MAXSCORE
          gotdir = 1;
#endif
          parents[j - 1] = j;
          add_to_chain(j - 1, chain, cstart, cend++, sf->score);
        }
        if(x < game.ncols  && parents[j] != j + 1 && parents[j + 1] == -1
              && XY(x + 1, y) != WALL && XY(x + 1, y) != LETHAL)
        {
#if USE_MAXSCORE
          gotdir = 1;
#endif
          parents[j + 1] = j;
          add_to_chain(j + 1, chain, cstart, cend++, sf->score);
        }
        if(y > 1 && parents[j] != j - game.ncols && parents[j - game.ncols] ==
              -1 && XY(x, y - 1) != WALL && XY(x, y - 1) != LETHAL)
        {
#if USE_MAXSCORE
          gotdir = 1;
#endif
          parents[j - game.ncols] = j;
          add_to_chain(j - game.ncols, chain, cstart, cend++, sf->score);
        }
        if(y < game.nrows  && parents[j] != j + game.ncols && parents[j +
              game.ncols] == -1 && XY(x, y + 1) != WALL && XY(x, y + 1) !=
              LETHAL)
        {
#if USE_MAXSCORE
          gotdir = 1;
#endif
          parents[j + game.ncols] = j;
          add_to_chain(j + game.ncols, chain, cstart, cend++, sf->score);
        }
#if USE_MAXSCORE
        if(!gotdir)
        {
          int maxsc;
          // Dead end.
          maxsc = sf->score;
          sf->maxscore = sf->score;
          while(parents[j] != -1)
          {
            if(send_now)
              break;
            
            j = parents[j];
            sf = &game.score_field[j];
            if(sf->maxscore < maxsc)
              sf->maxscore = maxsc;
            else
              break;
          }
        }
#endif
      }
      area++;
    }
    if(++i == game.ncols * game.nrows)
      i = 0;
  } while(i != si);
  
  free(chain);
  free(parents);
}

int find_closest_tile(char type)
{
  int res = -1, resscore = 0x7fffffff;
  int x, y;
  
  for(y = 1; y <= game.nrows; y++)
    for(x = 1; x <= game.ncols; x++)
    {
      if(XY(x, y) == type && SXY(x, y).score < resscore && SXY(x, y).area <= 1
            && SXY(x, y).dir)
      {
        resscore = SXY(x, y).score;
        res = (y - 1) * game.ncols + x - 1;
      }
    }
  
  return res;
}

void calculate_package_scores(void)
{
  package_t *pkg;
  int w;
  int *dsts, *srcs;
  
  dsts = calloc(game.nrows * game.ncols, sizeof(int));
  srcs = calloc(game.nrows * game.ncols, sizeof(int));
  
  for(pkg = packages; pkg; pkg = pkg->next)
  {
    if(pkg->rabid)
    {
      pkg->score = 0;
      continue;
    }
    if(pkg->xgoal != -1)
      dsts[(pkg->ygoal - 1) * game.ncols + pkg->xgoal - 1]++;
    srcs[(pkg->y - 1) * game.ncols + pkg->x - 1]++;
    if(pkg->w == -1)
      w = rintf(pMWorth);
    else
      w = pkg->w;
    if(pkg->carrier == me->n)
    {
      if(SXY(pkg->xgoal, pkg->ygoal).area <= 1)
        pkg->score = w / (float)SXY(pkg->xgoal, pkg->ygoal).score;
      else
        pkg->score = 0;
    }
    else if(pkg->carrier != -1)
      pkg->score = 0;
    else
    {
      if(pkg->xgoal != -1 && SXY(pkg->xgoal, pkg->ygoal).area <= 1 &&
            SXY(pkg->x, pkg->y).area <= 1)
        pkg->score = w / (float)(2 * SXY(pkg->x, pkg->y).score + SXY(pkg->xgoal,
              pkg->ygoal).score);
      else
        pkg->score = w / 900000.;
    }
  }
  
  for(pkg = packages; pkg; pkg = pkg->next)
  {
    w = srcs[(pkg->y - 1) * game.ncols + pkg->x - 1];
    if(pkg->xgoal != -1)
      w += dsts[(pkg->ygoal - 1) * game.ncols + pkg->xgoal - 1] - 1;
    pkg->score *= w;
  }
  
  free(dsts);
  free(srcs);
}

char *think_drop(void)
{
  int gotone;
  package_t *pkg, *found;
  static char buf[BUFSIZE];
  int minfpkg;
  
  sprintf(buf, "1 Drop");
  gotone = 0;
  found = NULL;
  minfpkg = 0x7fffffff;
  for(pkg = packages; pkg; pkg = pkg->next)
  {
    if(pkg->carrier == me->n)
    {
      if(pkg->xgoal == me->x && pkg->ygoal == me->y)
      {
        sprintf(&buf[strlen(buf)], " %d", pkg->n);
        gotone = 1;
      }
      else if(!gotone)
      {
        if(SXY(pkg->xgoal, pkg->ygoal).score < minfpkg &&
              SXY(pkg->xgoal, pkg->ygoal).area <= 1 &&
              SXY(pkg->xgoal, pkg->ygoal).dir)
        {
          found = pkg;
          minfpkg = SXY(pkg->xgoal, pkg->ygoal).score;
        }
      }
    }
  }
  if(gotone)
  {
    strcat(buf, "\n");
    return buf;
  }
  if(found)
  {
    sprintf(buf, "%d Move %c\n", force, SXY(found->xgoal, found->ygoal).dir);
    return buf;
  }
  
  return NULL;
}

char *think_pick(void)
{
  static char buf[BUFSIZE];
  package_t *pkg, *found;
  int gotone;
  player_t *pl;
  float maxsc, maxacsc;
  
  strcpy(buf, "1 Pick");
  gotone = 0;
  found = (package_t*)1;
  maxacsc = 0x7fffffff;
  while(found && !send_now)
  {
    if(strlen(buf) > 100)
      break;
    found = NULL;
    maxsc = -1;
    for(pkg = packages; pkg; pkg = pkg->next)
    {
      if(strlen(buf) > 100)
        break;
      if(pkg->rabid && pkg->rabid < 8)
        continue;
      if(pkg->x != me->x || pkg->y != me->y || pkg->w == -1
            || pkg->xgoal == -1)
        continue;
      
      if(pkg->score == maxacsc)
      {
        sprintf(&buf[strlen(buf)], " %d", pkg->n);
        gotone = 1;
      }
      else if(pkg->score > maxsc && pkg->score < maxacsc)
      {
        found = pkg;
        maxsc = pkg->score;
        gotone = 1;
      }
    }
    maxacsc = maxsc;
  }
  if(gotone)
  {
    for(pl = players; pl; pl = pl->next)
    {
      if(pl->x == me->x && (pl->y == me->y + 1 || pl->y == me->y - 1))
        break;
      if(pl->y == me->y && (pl->x == me->x + 1 || pl->x == me->x - 1))
        break;
    }
    if(pl)
    {
      for(pkg = packages; pkg; pkg = pkg->next)
        if(pkg->x == me->x && pkg->y == me->y && !pkg->rabid)
          pkg->rabid = 10;
      if(XY(me->x, me->y) == DANGEROUS)
      {
        if(me->x > 1 && me->x < game.ncols && XY(me->x - 1, me->y) == LETHAL &&
              XY(me->x + 1, me->y) != WALL && XY(me->x + 1, me->y) != LETHAL)
        {
          force = 1;
          strcpy(buf, "1 Move E");
        }
        else if(me->x > 1 && me->x < game.ncols && XY(me->x + 1, me->y) ==
              LETHAL && XY(me->x - 1, me->y) != WALL && XY(me->x - 1, me->y) !=
              LETHAL)
        {
          force = 1;
          strcpy(buf, "1 Move W");
        }
        else if(me->y > 1 && me->y < game.nrows && XY(me->x, me->y + 1) ==
              LETHAL && XY(me->x, me->y - 1) != WALL && XY(me->x, me->y - 1) !=
              LETHAL)
        {
          force = 1;
          strcpy(buf, "1 Move S");
        }
        else if(me->y > 1 && me->y < game.nrows && XY(me->x, me->y - 1) ==
              LETHAL && XY(me->x, me->y + 1) != WALL && XY(me->x, me->y + 1) !=
              LETHAL)
        {
          force = 1;
          strcpy(buf, "1 Move N");
        }
      }
    }
    strcat(buf, "\n");
    return buf;
  }
  
  return NULL;
}

char *think_move(void)
{
  int i;
  static char buf[BUFSIZE];
  float maxsc;
  package_t *pkg;
  
  i = find_closest_tile(BASE);
  if(i >= 0)
    maxsc = bMWorth / (float)game.score_field[i].score;
  else
    maxsc = 0;
  for(pkg = packages; pkg; pkg = pkg->next)
  {
    if(pkg->rabid && pkg->rabid < 8)
      continue;
    if(pkg->score > maxsc)
    {
      i = (pkg->y - 1) * game.ncols + pkg->x - 1;
      maxsc = pkg->score;
    }
  }
  if(i >= 0)
  {
    sprintf(buf, "%d Move %c\n", force, game.score_field[i].dir);
    return buf;
  }
  
  return NULL;
}

char *think_alone(void)
{
  char *next_line;
  int i;
  static char buf[BUFSIZE];
  
  if(!packages)
  {
    if(score_field_each_turn)
    {
      i = find_closest_tile(BASE);
      
      if(i < 0)
        return "1 Drop\n";
      
      sprintf(buf, "%d Move %c\n", force, game.score_field[i].dir);
      return buf;
    }
  }
  
  if((next_line = think_drop()))
    return next_line;
  
  if((next_line = think_pick()))
    return next_line;
  
  if((next_line = think_move()))
    return next_line;
  
  return "1 Drop\n";
}

char *think_kill(void)
{
  int x, y;
  player_t *pl;
  static char buf[BUFSIZE];
  
  for(pl = players; pl; pl = pl->next)
  {
    if(send_now)
      return NULL;
    x = y = 0;
    if(pl->x == me->x)
    {
      if(pl->y == me->y - 1)
        y = -1;
      else if(pl->y == me->y + 1)
        y = 1;
    }
    else if(pl->y == me->y)
    {
      if(pl->x == me->x - 1)
        x = -1;
      else if(pl->x == me->x + 19)
        x = 1;
    }
    if(x || y)
    {
      if(XY(me->x + 2 * x, me->y + 2 * y) == LETHAL)
      {
        const char dirs[] = "SW0EN";
        
        force = 20;
        sprintf(buf, "20 Move %c\n", dirs[2 + 2 * y + x]);
        return buf;
      }
    }
    if(abs(pl->x - me->x) == 1 && abs(pl->y - me->y) == 1)
    {
      x = pl->x - me->x;
      y = pl->y - me->y;
      
      if(XY(me->x, pl->y) == DANGEROUS && XY(me->x, me->y + 2 * y) ==
            LETHAL && pl->lx != me->x)
      {
        const char dirs[] = "S0N";
        
        force = -2;
        sprintf(buf, "-2 Move %c\n", dirs[1 + y]);
        return buf;
      }
      if(XY(pl->x, me->y) == DANGEROUS && XY(me->x + 2 * x, me->y) ==
            LETHAL && pl->ly != me->y)
      {
        const char dirs[] = "W0E";
        
        force = -2;
        sprintf(buf, "-2 Move %c\n", dirs[1 + x]);
        return buf;
      }
    }
  }
  return NULL;
}

char *think(void)
{
  char *next_line;
  
  if(!me)
    return "1 Drop\n";
  if(XY(me->x, me->y) == BASE)
  {
    if(me->x > 1 && XY(me->x - 1, me->y) == LETHAL)
      XY(me->x, me->y) = DANGEROUS;
    else if(me->x < game.ncols && XY(me->x + 1, me->y) == LETHAL)
      XY(me->x, me->y) = DANGEROUS;
    else if(me->y > 1 && XY(me->x, me->y - 1) == LETHAL)
      XY(me->x, me->y) = DANGEROUS;
    else if(me->y < game.nrows && XY(me->x, me->y + 1) == LETHAL)
      XY(me->x, me->y) = DANGEROUS;
    else
      XY(me->x, me->y) = NORMAL;
  }
  
  calculate_package_scores();
  
  if(all_alone)
    return think_alone();
  
  if((next_line = think_kill()))
    return next_line;
  
  //calculate_player_destinations();
  
  return think_alone();
}

char *post_think(char *line)
{
  static char buf[BUFSIZE];
  static char tbuf[BUFSIZE];
  int x, y;
  player_t *pl;
  package_t *pkg;
  
  want_go = 0;
  sscanf(line, "%*d Move %c", &want_go);
  
  if(send_now)
    return line;
  
  if(want_go)
  {
    x = y = 0;
    switch(want_go)
    {
    case 'N':
      y = 1;
      break;
    case 'S':
      y = -1;
      break;
    case 'E':
      x = 1;
      break;
    case 'W':
      x = -1;
      break;
    }
    for(pl = players; pl; pl = pl->next)
    {
      if(pl == me)
        continue;
      if(pl->x == me->x + x - 1 && pl->y == me->y + y &&
            XY(me->x + x + 1, me->y + y) == LETHAL)
      {
        force = -20;
        break;
      }
      if(pl->x == me->x + x + 1 && pl->y == me->y + y &&
            XY(me->x + x - 1, me->y + y) == LETHAL)
      {
        force = -20;
        break;
      }
      if(pl->x == me->x + x && pl->y == me->y + y - 1 &&
            XY(me->x + x, me->y + y + 1) == LETHAL)
      {
        force = -20;
        break;
      }
      if(pl->x == me->x + x && pl->y == me->y + y + 1 &&
            XY(me->x + x, me->y + y - 1) == LETHAL)
      {
        force = -20;
        break;
      }
      if(pl->x == me->x + x && pl->y == me->y + y)
      {
        force += 10;
        for(pkg = packages; pkg; pkg = pkg->next)
          if(pkg->carrier == me->n)
            force += 14;
        if(XY(me->x - 2 * x, me->y + 2 * y) == WALL)
        {
          // Can't do this... bolt!
          force = 1;
          want_go = 0;
          while(!want_go && !send_now)
          {
            switch(rand() % 4)
            {
            case 0:
              if(y != 1 && XY(me->x, me->y + 1) != WALL && XY(me->x, me->y + 1)
                    != LETHAL)
                want_go = 'N';
              break;
            case 1:
              if(y != -1 && XY(me->x, me->y - 1) != WALL && XY(me->x, me->y - 1)
                    != LETHAL)
                want_go = 'S';
              break;
            case 2:
              if(x != 1 && XY(me->x + 1, me->y) != WALL && XY(me->x + 1, me->y)
                    != LETHAL)
                want_go = 'E';
              break;
            case 3:
              if(x != -1 && XY(me->x - 1, me->y) != WALL && XY(me->x - 1, me->y)
                    != LETHAL)
                want_go = 'W';
              break;
            }
          }
          if(want_go)
            sprintf(buf, "1 Move %c\n", want_go);
        }
        else if(XY(me->x + 2 * x, me->y + 2 * y) == LETHAL)
          force += 20;
        if(XY(me->x - x, me->y - y) == LETHAL)
          force += 40;
      }
      if(force > 100)
        force = -1;
    }
    sscanf(line, "%*d %[^\n]", tbuf);
    sprintf(buf, "%d %s\n", force, tbuf);
    return buf;
  }
  return line;
}

int initialize(void)
{
  const char initSend[] = "Player\n";
  static char read_line[BUFSIZE];
  int x, y;
  int men;
  struct rusage before, after;
  
  put_line(initSend);
  
  if(!get_line(read_line))
  {
    fprintf(stderr, "Failed to get board dimensions.\n");
    return -1;
  }
  if(sscanf(read_line, "%d %d", &game.ncols, &game.nrows) != 2)
  {
    fprintf(stderr, "Board dimensions not matching expected format.\n");
    return -1;
  }
  if(game.ncols > sizeof(read_line) - 1)
  {
    fprintf(stderr, "Number of columns higher that the maximum supported %d.\n",
          (int)sizeof(read_line) - 1);
    return -1;
  }
  
  game.field = malloc(game.ncols * game.nrows);
  
  for(y = 0; y < game.nrows; y++)
  {
    if(!get_line(read_line))
    {
      fprintf(stderr, "Failed to get board row %d.\n", y + 1);
      return -1;
    }
    if(strlen(read_line) - 1 != game.ncols)
    {
      fprintf(stderr, "Board row %d: Expected %d characters, got %d.\n", y + 1,
            game.ncols, strlen(read_line) - 1);
      return -1;
    }
    memcpy(&game.field[y * game.ncols], read_line, game.ncols);
  }
  
  if(!get_line(read_line))
  {
    fprintf(stderr, "Failed to get player initialization.\n");
    return -1;
  }
  if(sscanf(read_line, "%d %d %d", &men, &max_carry, &money) != 3)
  {
    fprintf(stderr, "Player initialization not matching expected format.\n");
    return -1;
  }
  me = find_player(men);
  if(!me)
  {
    perror("malloc");
    return -1;
  }
  
  for(y = 1; y <= game.nrows; y++)
    for(x = 1; x <= game.ncols; x++)
    {
      if(XY(x, y) == LETHAL)
      {
        if(x > 1 && XY(x - 1, y) == NORMAL)
          XY(x - 1, y) = DANGEROUS;
        if(x < game.ncols && XY(x + 1, y) == NORMAL)
          XY(x + 1, y) = DANGEROUS;
        if(y > 1 && XY(x, y - 1) == NORMAL)
          XY(x, y - 1) = DANGEROUS;
        if(y < game.nrows && XY(x, y + 1) == NORMAL)
          XY(x, y + 1) = DANGEROUS;
      }
    }
  
  send_now = 0;
  getrusage(0, &before);
  setitimer(ITIMER_PROF, &timer, NULL);
  make_score_field(0, 0);
  getrusage(0, &after);
  setitimer(ITIMER_PROF, &timer0, NULL);
  if(send_now)
    printf("Warning: score field canceled since it took too long.\n");
  else
  {
    if(after.ru_utime.tv_sec > before.ru_utime.tv_sec)
      after.ru_utime.tv_usec += 1000000000;
    printf("Score field finished in %u microseconds.\n",
          after.ru_utime.tv_usec - before.ru_utime.tv_usec);
    if(after.ru_utime.tv_usec - before.ru_utime.tv_usec <= MAX_SCORE_TIME)
      score_field_each_turn = 1;
  }
  
  if(verbose >= 2)
  {
    for(y = game.nrows; y >= 1; y--)
    {
      for(x = 1; x <= game.ncols; x++)
        printf("%c", SXY(x, y).area ? XY(x, y) : ' ');
      printf("\n");
    }
    for(y = game.nrows; y >= 1; y--)
    {
      for(x = 1; x <= game.ncols; x++)
        printf("%d ", SXY(x, y).area);
      printf("\n");
    }
  }
  
  if(!get_line(read_line))
  {
    fprintf(stderr, "Failed to get initial player positions.\n");
    return -1;
  }
  process_event_line(read_line);
  
  return 0;
}

void usage(char *prog)
{
  fprintf(stderr, "Syntax: %s [-v [-v [-v]]] <host> <port>\n", prog);
  exit(1);
}

int main(int argc, char *argv[])
{
  char *next_line;
  static char read_line[BUFSIZE];
  package_t *pkg;
  int x, y;
  player_t *pl;
  
  next_line = argv[0];
  
  while(argv[1][0] == '-')
  {
    switch(argv[1][1])
    {
    case 'v':
      verbose++;
      break;
    default:
      usage(next_line);
    }
    argv++;
    argc--;
  }
  if(argc != 3)
    usage(next_line);
  
  sock = open_conn(argv[1], argv[2]);
  if(!sock)
    return 1;
  
  if(initialize())
    return 1;
  
  signal(SIGPROF, sigprof);
  
  if(verbose)
    printf("\x1b[2J");
  
  force = 1;
  while(get_line(read_line))
  {
    send_now = 0;
    
    setitimer(ITIMER_PROF, &timer, NULL);
    
    if(verbose >= 2)
    {
      printf("\x1b[1;1H");
      for(y = game.nrows; y >= 1; y--)
      {
        for(x = 1; x <= game.ncols; x++)
          printf("%c", XY(x, y));
        printf("\n");
      }
      for(pkg = packages; pkg; pkg = pkg->next)
        printf("\x1b[%d;%dHp", game.nrows + 1 - pkg->y, pkg->x);
      
      for(pl = players; pl; pl = pl->next)
      {
        for(pkg = packages; pkg; pkg = pkg->next)
        {
          if(pkg->x == pl->x && pkg->y == pl->y)
          {
            if(pl == me)
              printf("\x1b[%d;%dHX", game.nrows + 1 - pl->y, pl->x);
            else
              printf("\x1b[%d;%dHE", game.nrows + 1 - pl->y, pl->x);
            break;
          }
        }
        if(!pkg)
        {
          if(pl == me)
            printf("\x1b[%d;%dHx", game.nrows + 1 - pl->y, pl->x);
          else
            printf("\x1b[%d;%dHe", game.nrows + 1 - pl->y, pl->x);
        }
      }
      printf("\x1b[%d;1H", game.nrows + 1);
    }
    if(verbose >= 3)
      printf("\x1b[0J%s", read_line);
    
    if(score_field_each_turn && me)
      make_score_field(me->x, me->y);
  
    process_package_line(read_line);
    
    if(verbose && me)
      printf("X %d Y %d score %d\x1b[0K\n", me->x, me->y, me->score);
    
    next_line = think();
    
    setitimer(ITIMER_PROF, &timer0, NULL);
    
    next_line = post_think(next_line);
    if(verbose)
      printf("\x1b[0K%c %s", send_now ? '!' : '>', next_line);
    put_line(next_line);
    
    if(!get_line(read_line))
      break;
    
    if(verbose)
      printf("\x1b[0K%s", read_line);
    process_event_line(read_line);
  }
  return 0;
}
